{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Practice №4. Solving the problem of video analytics, including detection, recognition and tracking of objects in the video\n",
    "\n",
    "In this practice, we propose a solution to the problem of counting vehicles of different classes on video and the development of an appropriate application using OpenVINO.\n",
    "\n",
    "To enable OpenVINO support in Jupyter Notebook please run a command line, activate the virtual environment (if any) in the command line and run the __setupvars__ script to activate OpenVINO, and then start Jupyter Notebook.\n",
    "\n",
    "For Windows:\n",
    "\n",
    "```bash\n",
    "<virtual_enviroment>/Scripts/activate.bat\n",
    "\"C:\\Program Files (x86)\\Intel\\openvino_2021\\bin\\setupvars.bat\"\n",
    "jupyter notebook\n",
    "```\n",
    "\n",
    "For Linux:\n",
    "```bash\n",
    "source <virtual_enviroment>/bin/activate\n",
    "source /opt/intel/openvino_2021/bin/setupvars.sh\n",
    "jupyter notebook\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Checking that OpenVINO is installed and connected\n",
    "import openvino"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Problem statement of counting vehicles on video\n",
    "\n",
    "In this practice, the problem of counting vehicles of different classes on the video from the crossroad camera is considered. The input is a sequence of video frames. It is required to count the number of vehicles of each category (“car”, “bus”, “motorbike” and others) that have already been observed in the video by the current moment."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Algorithm for counting vehicles of different classes on video\n",
    "The base algorithm for counting vehicles of different classes on video based on the object detection and tracking consists of the several steps performed processing the next received frame.\n",
    "\n",
    "1.\tObject detection. To detect objects, any deep model capable of detecting vehicles of different classes (“car”, “bus”, “train”, and “motorbike”) can be used. It is required to filter out objects that do not belong to the set of vehicle classes. Filtering can be performed at the detection stage or at the stage of calculating statistics. It is also necessary to remember that the detection algorithm for the same object on different frames can predict a different class. In our solution, the class of the object is the class predicted at the last processed frame. For vehicle detection, the SSD300 model and InferenceEngineDetector class can be used.\n",
    "\n",
    "2.\tObject tracking. You can use tracking-by-matching implementation of the tracker from the practice “Tracking objects on video using deep neural networks”. \n",
    "\n",
    "3.\tCalculating statistics for vehicles of different classes, which were observed from the beginning of the video to the current frame. To calculate statistics, it is required to create counters for each class of vehicles and to iterate over all the tracks, increasing the counters of the corresponding vehicle class.\n",
    "\n",
    "4.\tDisplaying frame with detected vehicles and their tracks, displaying the collected statistic information on the screen.\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import cv2\n",
    "import numpy as np\n",
    "from openvino.inference_engine import IECore\n",
    "from scipy.optimize import linear_sum_assignment\n",
    "import math\n",
    "from collections import namedtuple\n",
    "\n",
    "class InferenceEngineDetector:\n",
    "    def __init__(self, configPath=None, weightsPath=None,\n",
    "            device='CPU', extension=None, classesPath=None):\n",
    "        self.ie = IECore()\n",
    "        self.net = self.ie.read_network(model=configPath, weights=weightsPath)\n",
    "        if extension:\n",
    "            self.ie.add_extension(extension, 'CPU')\n",
    "        self.exec_net = self.ie.load_network(network=self.net,\n",
    "                                             device_name=device)\n",
    "        if classesPath:\n",
    "            self.classes = [line.rstrip('\\n') for line in open(classesPath)]\n",
    "        return\n",
    "\n",
    "    def _prepare_image(self, image, h, w):\n",
    "        if image.shape[:-1] != (h, w):\n",
    "            image = cv2.resize(image, (w, h))\n",
    "        image = image.transpose((2, 0, 1))\n",
    "        return image\n",
    "\n",
    "    def detect(self, image):\n",
    "        input_blob = next(iter(self.net.input_info))\n",
    "        out_blob = next(iter(self.net.outputs))\n",
    "        n, c, h, w = self.net.input_info[input_blob].input_data.shape\n",
    "        blob = self._prepare_image(image, h, w)\n",
    "        output = self.exec_net.infer(inputs={input_blob: blob})\n",
    "        detection = output[out_blob]\n",
    "        return detection\n",
    "\n",
    "    def draw_detection(self, detections, img, confidence=0.5, draw_text=True):\n",
    "        (h, w) = img.shape[:2]\n",
    "        for i in range(0, detections.shape[2]):\n",
    "            conf = detections[0, 0, i, 2]\n",
    "            if conf > confidence:\n",
    "                box = detections[0, 0, i, 3:7] * np.array([w, h, w, h])\n",
    "                (startX, startY, endX, endY) = box.astype(\"int\")\n",
    "                try:\n",
    "                    text = \"Class {} {:.2f}%\".format(\n",
    "                        self.classes[int(detections[0, 0, i, 1])], conf * 100)\n",
    "                except:\n",
    "                    text = \"Class {} {:.2f}%\".format(\n",
    "                        detections[0, 0, i, 1], conf * 100)\n",
    "                y = startY - 10 if startY - 10 > 10 else startY + 10\n",
    "                cv2.rectangle(img, (startX, startY), (endX, endY),\n",
    "                              (0, 255, 0), 2)\n",
    "                if draw_text:\n",
    "                    cv2.putText(img, text, (startX, y),\n",
    "                                cv2.FONT_HERSHEY_COMPLEX, 0.45, (0, 0, 255), 1)\n",
    "        return img\n",
    "\n",
    "DetectedObject = namedtuple(\n",
    "    'DetectedObject', \n",
    "    ['confidence', 'frame_idx', 'object_id', 'timestamp', \n",
    "     'class_id', 'x_left', 'y_bottom', 'x_right', 'y_top',])    \n",
    "    \n",
    "class Tracklet():\n",
    "    def __init__(self, detection=[]):\n",
    "        self._trackedObjects = []\n",
    "        self._trackedObjects.append(detection)\n",
    "    def __len__(self):\n",
    "        return len(self._trackedObjects)\n",
    "    def __getitem__(self, position):\n",
    "        return self._trackedObjects[position]\n",
    "    def add_new_detection(self, detection):\n",
    "        self._trackedObjects.append(detection)\n",
    "    \n",
    "    \n",
    "class MatchingTracker:\n",
    "    \n",
    "    def _affinity(self, obj1, obj2):\n",
    "        shp_aff = self._shape_affinity(obj1, obj2, self._shape_affinity_weight)\n",
    "        mot_aff = self._place_affinity(obj1, obj2, self._place_affinity_weight)\n",
    "        return shp_aff * mot_aff      \n",
    "    \n",
    "    def _shape_affinity(self, obj1, obj2, weight):\n",
    "        obj1_width = obj1.x_right-obj1.x_left\n",
    "        obj2_width = obj2.x_right-obj2.x_left\n",
    "        obj1_height = obj1.y_top - obj1.y_bottom\n",
    "        obj2_height = obj2.y_top - obj2.y_bottom\n",
    "        w_dist = abs(obj1_width - obj2_width) / (obj1_width + obj2_width)\n",
    "        h_dist = abs(obj1_height - obj2_height) / (obj1_height + obj2_height)\n",
    "        return math.exp(-weight * (w_dist + h_dist))\n",
    "    \n",
    "    def _place_affinity(self, o1, o2, weight):\n",
    "        obj1_x = (o1.x_left + o1.x_right) * 0.5\n",
    "        obj1_y = (o1.y_top + o1.y_bottom) * 0.5\n",
    "        obj2_x = (o2.x_left + o2.x_right) * 0.5\n",
    "        obj2_y = (o2.y_top + o2.y_bottom) * 0.5\n",
    "        obj2_width = o2.x_right-o2.x_left\n",
    "        obj2_height = o2.y_top - o2.y_bottom\n",
    "        x_dist = ((obj1_x - obj2_x) ** 2) / (obj2_width ** 2)\n",
    "        y_dist = ((obj1_y - obj2_y) ** 2) / (obj2_height ** 2)\n",
    "        return math.exp(-weight * (x_dist + y_dist))\n",
    "    \n",
    "    def _compute_dissimilarity_matrix(self, tracks, detections):\n",
    "        size = max(len(tracks), len(detections))\n",
    "        diss_mat = np.zeros(shape=(size, size), dtype=float)\n",
    "        for i in range(len(tracks)):\n",
    "            for j in range(len(detections)):\n",
    "                diss_mat[i, j] = 1.0 - \\\n",
    "                    self._affinity(tracks[i][-1], detections[j])\n",
    "        return diss_mat\n",
    "    \n",
    "    def _solve_assignment_problem(self, tracks, detections):\n",
    "        dissimilarity_mat = self._compute_dissimilarity_matrix(\n",
    "            tracks, detections)\n",
    "        row_ind, col_ind = linear_sum_assignment(dissimilarity_mat)\n",
    "        return row_ind, col_ind\n",
    "\n",
    "    def filter_detections(self, detect_mat, threshold=0.5):\n",
    "        detect_mat = detect_mat[0, 0, :, :]\n",
    "        detections = []\n",
    "        for i in range(detect_mat.shape[0]):\n",
    "            # Parse one string in ie_detection_output\n",
    "            conf = detect_mat[i, 2]\n",
    "            if conf > threshold:\n",
    "                detection = DetectedObject(\n",
    "                    detect_mat[i, 2],\n",
    "                    -1,\n",
    "                    -1,\n",
    "                    -1,\n",
    "                    detect_mat[i, 1],\n",
    "                    detect_mat[i, 3],\n",
    "                    detect_mat[i, 4],\n",
    "                    detect_mat[i, 5],\n",
    "                    detect_mat[i, 6])\n",
    "                detections.append(detection)\n",
    "        return detections  \n",
    "\n",
    "    def add_new_track(self, detection):\n",
    "        track = Tracklet(detection)\n",
    "        self._tracks.append(track)\n",
    "        \n",
    "    def process_new_frame(self, frame, detections, timestamp):\n",
    "        if self._tracks and detections:\n",
    "            row_indexes, col_indexes = self._solve_assignment_problem(\n",
    "                self._tracks, detections)\n",
    "            # For each assignment\n",
    "            for i in range(len(row_indexes)):\n",
    "                row_id = row_indexes[i]\n",
    "                col_id = col_indexes[i]\n",
    "                # If we find existing track and existing detection\n",
    "                if row_id < len(self._tracks) and col_id < len(detections):\n",
    "                    # Add detection to track if objects are close\n",
    "                    dist = 1.0 - \\\n",
    "                        self._affinity(\n",
    "                            self._tracks[row_id][-1], detections[col_id])\n",
    "                    if dist < self._dist_weight:\n",
    "                        self._tracks[row_id].add_new_detection(\n",
    "                            detections[col_id])\n",
    "                elif col_id < len(detections):\n",
    "                    self.add_new_track(detections[col_id])\n",
    "        else:\n",
    "            # Add new tracks\n",
    "            for id, detection in enumerate(detections):\n",
    "                self.add_new_track(detection)\n",
    "        return\n",
    "\n",
    "    def __init__(self, dist_weight=0.5, shape_affinity_weight=0.5,\n",
    "            place_affinity_weight=0.7):\n",
    "        self._tracks = []\n",
    "        self._tracks_counter = 0\n",
    "        self._dist_weight = dist_weight\n",
    "        self._shape_affinity_weight = shape_affinity_weight\n",
    "        self._place_affinity_weight = place_affinity_weight\n",
    "        return\n",
    "    \n",
    "    def draw_active_tracks(self, image):\n",
    "        w, h = image.shape[:2]\n",
    "        for i, track in enumerate(self._tracks):\n",
    "            # Draw one track from segments\n",
    "            for i in range(len(track)-1):\n",
    "                cv2.line(img=image,\n",
    "                    pt1=(int((track[i].x_left + track[i].x_right) * h) // 2,\n",
    "                         int((track[i].y_bottom + track[i].y_top) * w) // 2),\n",
    "                    pt2=(int((track[i+1].x_left + track[i+1].x_right) * h) // 2,\n",
    "                         int((track[i+1].y_bottom + track[i+1].y_top) * w) // 2),\n",
    "                    color=(0, 255, 0), thickness=3)\n",
    "        return image"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Developing the application for counting vehicles of different classes on video\n",
    "\n",
    "It is assumed that the InferenceEngineDetector and MatchingTracker modules, which contain the implementation of algorithms for detecting and tracking objects, have already been developed during the previous practical work. Let's take a closer look at the Videoanalytics class. The class contains the following methods:\n",
    "\n",
    "- _init_ is a class constructor. It loads a list of vehicle classes to be counted, and the names of these classes to display on the screen.\n",
    "- count_objects_per_classes is a method that counts vehicles of each class.\n",
    "- draw_videoanalytics is a method for displaying the vehicle counting results on the video.\n",
    "\n",
    "The constructor implementation is trivial, so it is proposed to develop it by yourself. The implementation of other methods will be described below"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import cv2\n",
    "import numpy as np\n",
    "\n",
    "\n",
    "class Videoanalytics():\n",
    "    def __init__(self, classIds, classesPath=None):\n",
    "        pass\n",
    "    def count_objects_per_classes(self, tracker):\n",
    "        pass\n",
    "    def draw_videoanalytics(self, frame, tracker):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "__Counting vehicles of different classes on video__\n",
    "\n",
    "Consider the implementation of the count_objects_per_classes method for counting the number of vehicles of each class. This method iterates over all tracks obtained as a result of tracking, and counts the number of objects of each class."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import cv2\n",
    "import numpy as np\n",
    "\n",
    "class Videoanalytics():\n",
    "\n",
    "    def draw_videoanalytics(self, frame, tracker):\n",
    "        pass\n",
    "\n",
    "# Constructor\n",
    "    def __init__(self, classIds, classesPath=None):\n",
    "        self.classIds = classIds\n",
    "        if classesPath:\n",
    "            self.classes = [line.rstrip('\\n') for line in open(classesPath)]\n",
    "        return\n",
    "\n",
    "# Counting objects\n",
    "    def count_objects_per_classes(self, tracker):\n",
    "        results = {}\n",
    "        for tracklet in tracker._tracks:\n",
    "            # Get object class from tracklet\n",
    "            classid = tracklet[-1].class_id\n",
    "            # Add +1 to counter for object\n",
    "            if classid in self.classIds:\n",
    "                if classid in results:\n",
    "                    results[classid] += 1\n",
    "                else:\n",
    "                    results[classid] = 1\n",
    "        return results"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "__Displaying of vehicle counting results__\n",
    "\n",
    "The draw_videoanalytics method displays the results of counting vehicles. It consists of three parts.\n",
    "\n",
    "1.\tDisplaying detected objects of vehicle classes. It is required to iterate over all tracks.\n",
    "\n",
    "    1.1.\tSearching for the last bounding box corresponding to the object location in the track.\n",
    "\n",
    "    1.2.\tIf the object class is in the list of vehicle classes and the object is found in the current frame, then displaying the bounding box.\n",
    "\n",
    "\n",
    "2.\tDisplaying the tracks of vehicles. It is supposed to iterate over all tracks.\n",
    "\n",
    "    2.1.\tSearching for the last bounding box corresponding to the object location in the track.\n",
    "\n",
    "    2.2.\tIf the object class is in the list of vehicle classes and the object is found in the current frame, then iterate over all bounding box in the track, calculate their centers and draw line between neighbor frame centers.\n",
    "\n",
    "\n",
    "3.\tDisplaying statistics.\n",
    "\n",
    "    3.1.\tCount the number of vehicles of each class using the count_objects_per_classes method.\n",
    "\n",
    "    3.2.\tFor each vehicle class, create a line to display the following information: “Class <class_name>: <number_of_objects>”.\n",
    "\n",
    "    3.3.\tDisplaying lines with counting information in the upper left corner of the frame."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import cv2\n",
    "import numpy as np\n",
    "\n",
    "class Videoanalytics():\n",
    "    def __init__(self, classIds, classesPath=None):\n",
    "        self.classIds = classIds\n",
    "        if classesPath:\n",
    "            self.classes = [line.rstrip('\\n') for line in open(classesPath)]\n",
    "        else:\n",
    "            self.classes = None\n",
    "        return\n",
    "\n",
    "    def count_objects_per_classes(self, tracker):\n",
    "        results = {}\n",
    "        for tracklet in tracker._tracks:\n",
    "            # Get object class from tracklet\n",
    "            classid = tracklet[-1].class_id\n",
    "            # Add +1 to counter for object\n",
    "            if classid in self.classIds:\n",
    "                if classid in results:\n",
    "                    results[classid] += 1\n",
    "                else:\n",
    "                    results[classid] = 1\n",
    "        return results\n",
    "    \n",
    "# Draw videoanalytics\n",
    "    def draw_videoanalytics(self, frame, tracker):\n",
    "        (h, w) = frame.shape[:2]\n",
    "\n",
    "        # Draw detections of chosen classes\n",
    "        for tracklet in tracker._tracks:\n",
    "            # Get object class from tracklet\n",
    "            classid = int(tracklet[-1].class_id)\n",
    "            if classid in self.classIds:\n",
    "                # Draw bbox of the last detection\n",
    "                x_left = int(tracklet[-1].x_left * w)\n",
    "                y_bottom = int(tracklet[-1].y_bottom * h)\n",
    "                x_right = int(tracklet[-1].x_right * w)\n",
    "                y_top = int(tracklet[-1].y_top * h)\n",
    "                cv2.rectangle(frame, (x_left, y_bottom), (x_right, y_top),\n",
    "                    (0, 255, 0), 2)\n",
    "        \n",
    "        # Draw tracklets of chosen classes\n",
    "        for track in tracker._tracks:\n",
    "            # Get object class from tracklet\n",
    "            classid = int(track[-1].class_id)\n",
    "            if classid in self.classIds:\n",
    "            # Draw one tracklet from segments \n",
    "                for i in range(len(track)-1):\n",
    "                    x1 = int((track[i].x_left + track[i].x_right) * w) // 2\n",
    "                    y1 = int((track[i].y_bottom + track[i].y_top) * h) // 2\n",
    "                    x2 = int((track[i+1].x_left + track[i+1].x_right) * w) // 2\n",
    "                    y2 = int((track[i+1].y_bottom + track[i+1].y_top) * h) // 2\n",
    "                    cv2.line(img=frame, pt1=(x1,y1), pt2=(x2,y2), \n",
    "                        color=(0, 255, 0), thickness=3)\n",
    "\n",
    "        # Draw statistics\n",
    "        counts = self.count_objects_per_classes(tracker)\n",
    "        for i, elem in enumerate(counts):\n",
    "            if self.classes:\n",
    "                id = int(elem)\n",
    "                text = 'Class {}: {} objects'.format(\n",
    "                    self.classes[id], counts[elem])\n",
    "            else:\n",
    "                text = 'Class {}: {} objects'.format(elem, counts[elem])\n",
    "            text_pos = (0, i * 30 + 30)\n",
    "            cv2.putText(frame, text, text_pos,\n",
    "                        cv2.FONT_HERSHEY_COMPLEX, 1.0, (0, 255, 255), 2)\n",
    "        return frame\n",
    "\n",
    "    "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Implementing sample\n",
    "\n",
    "1.\tCreating an object of the InferenceEngineDetector class.\n",
    "\n",
    "2.\tCreating an object of the MatchingTracker class with empirically selected affinity weight parameters for the input video.\n",
    "\n",
    "3.\tCreating an object of the Videoanalytics class with the necessary parameters.\n",
    "\n",
    "4.\tLoading the video.\n",
    "\n",
    "5.\tPerforming actions for each video frame:\n",
    "\n",
    "    5.1.\tDetecting objects in the image.\n",
    "    \n",
    "    5.2.\tFiltering detections using the tracker.filter_detections method.\n",
    "    \n",
    "    5.3.\tTracking objects using the tracker.process_new_frame method.\n",
    "    \n",
    "    5.4.\tDisplaying detections and tracks of vehicles on the frame and displaying the statistics on the frame using the videoanalytics.draw_videoanalytics method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "model_path = 'public/ssd300/FP32/ssd300.xml'\n",
    "weights_path = 'public/ssd300/FP32/ssd300.bin'\n",
    "device = 'CPU' # 'CPU' or 'GPU'\n",
    "input_video = 'camera.mp4'\n",
    "classes_for_videoanalytics = [2,6,7,19]\n",
    "\n",
    "ie_detector = InferenceEngineDetector(configPath=model_path, weightsPath=weights_path, device=device) \n",
    "\n",
    "tracker = MatchingTracker()\n",
    "\n",
    "videoanalytics = Videoanalytics(classes_for_videoanalytics)\n",
    "\n",
    "# Open imput video\n",
    "cap = cv2.VideoCapture(input_video)\n",
    "\n",
    "# Setup output_video\n",
    "fps = cap.get(cv2.CAP_PROP_FPS)\n",
    "target_width = int(cap.get(3))\n",
    "target_height = int(cap.get(4))\n",
    "video_output_size = (target_width, target_height)\n",
    "fourcc = cv2.VideoWriter_fourcc(*'XVID')\n",
    "output_video = cv2.VideoWriter('output.mp4', fourcc, fps, video_output_size)\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "# Main loop\n",
    "timestamp = 0\n",
    "while(cap.isOpened()):\n",
    "    timestamp += 1\n",
    "    ret, frame = cap.read()\n",
    "\n",
    "    if ret == False:\n",
    "        break\n",
    "    \n",
    "    detections_mat = ie_detector.detect(frame)\n",
    "    detections = tracker.filter_detections(detections_mat, 0.5)\n",
    "\n",
    "    tracker.process_new_frame(frame, detections, timestamp)\n",
    "\n",
    "    result_image = videoanalytics.draw_videoanalytics(frame, tracker)\n",
    "\n",
    "    output_video.write(result_image)\n",
    "\n",
    "cap.release()\n",
    "output_video.release()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# play result video\n",
    "from IPython.display import Video\n",
    "output_video = 'output.mp4'\n",
    "Video(output_video, embed=True) "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "You can use the resulting Videoanalytics class paired with the MatchingTracker and InferenceEngineDetector classes in your own deep learning projects.\n",
    "\n",
    "\n",
    "### Дополнительные задания\n",
    "\n",
    "\n",
    "The developed sample contains the minimum required functionality. As additional tasks, it is proposed to provide support for the following features:\n",
    "\n",
    "1.\tThe scheme with storing information about all detected objects in tracks is not effective. It is proposed to filter non-vehicles immediately after detection and track vehicles.\n",
    "\n",
    "2.\tThe detector on different frames for the same object can predict various object classes. It is supposed to implement a scheme for determining the object class based on the full track information, not just the last frame. The easiest scheme is the voting one, when the object class is determined by the majority of the bounding boxes in the track.\n",
    "\n",
    "\n",
    "It is proposed to solve these tasks independently using the documentation and examples included in the OpenVINO Toolkit."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.7"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
